'use strict';

const { gl_Colors, gl_Shifts, gl_Shifts_map, gl_WidthShifts } = require('./render-utils.js');
const { vertexShader, fragmentShader } = require('./render-shader.js');

// const { ChangoItem } = require('./types.d.ts');

class WebGL2WaveRender {
    /**
     * 
     * @param {{
     * grid: HTMLDivElement,
     * view: HTMLDivElement,
     * values: HTMLDivElement
     * }} elements
     * @param {{
     * chango: Record<string, {
     *      kind: string,
     *      wave: Array<number | string>,
     *      lineVao: WebGLVertexArrayObject
     *      maskVao: WebGLVertexArrayObject
     * }>
     * view: Array<{ ref: string }>,
     * currentWires: Set<{
     *      kind: string,
     *      link: string,
     *      name: string,
     *      size: number,
     *      parent: object,
     *      type: string
     * }>,
     * time: number
     * }} globalLookup
     * @param {{
     * width: number,
     * height: number,
     * xScale: number,
     * xOffset: number,
     * yOffset: number,
     * yStep: number,
     * yDuty: number
     * }} pstate 
     * @param { Array } plugins
     */
    constructor(elements, globalLookup, pstate, plugins) {
        const canvas = document.createElement('canvas');
        elements.view.replaceChildren(canvas);

        this.elements = elements;
        this.globalLookup = globalLookup;
        this.canvas = canvas;
        this.pstate = pstate;
        this.plugins = plugins;

        const gl = canvas.getContext('webgl2', {
            premultipliedAlpha: false,
            alpha: true,
            antialias: false,
            depth: false
        });
        this.webglLocation = this.initProgram(gl);
        const { lineVerticesMap, maskVerticesMap } = this.makeVertex();
        this.lineVerticesMap = lineVerticesMap;
        this.maskVerticesMap = maskVerticesMap;

        this.initData();
        this.animationHandler = undefined;
    }

    /**
     * 
     * @param {WebGL2RenderingContext} gl 
     * @returns {{
     * colors: WebGLUniformLocation,
     * shifts: WebGLUniformLocation,
     * scale: WebGLUniformLocation,
     * offset: WebGLUniformLocation,
     * pos: number,
     * widthShifts: WebGLUniformLocation,
     * gl: WebGL2RenderingContext
     * }} 
    */
    initProgram(gl) {
        const program = gl.createProgram();
        gl.attachShader(program, vertexShader.make(gl));
        gl.attachShader(program, fragmentShader.make(gl));
        gl.linkProgram(program);
        gl.useProgram(program);
    
        const webglLocation = {
            colors: gl.getUniformLocation(program, 'colors'),
            shifts:  gl.getUniformLocation(program, 'shifts'),
            scale:  gl.getUniformLocation(program, 'scale'),
            offset: gl.getUniformLocation(program, 'offset'),
            pos: gl.getAttribLocation(program, 'pos'),
            widthShifts: gl.getUniformLocation(program, 'widthShifts'),
            gl
        };
    
        return webglLocation;
    }

    /**
     * 
     * @returns {{
     *  lineVerticesMap: Map<string, Uint32Array>,
     *  maskVerticesMap: Map<string, Uint32Array>
     * }}
     */
    makeVertex() {
        const globalLookup = this.globalLookup;
        const time = globalLookup.time;
        const lineVerticesMap = new Map();
        const maskVerticesMap = new Map();
        for (const id of Reflect.ownKeys(globalLookup.chango)) {
            const signalItem = globalLookup.chango[id];
            const { kind, wave } = signalItem;
            if (kind === 'bit') {
                const { lineVertices, maskVertices } = this.makeBitVertex(wave, time);
                lineVerticesMap.set(id, lineVertices);
                maskVerticesMap.set(id, maskVertices);
            } else if (kind === 'vec') {
                // const vertices = this.makeVecVertex(wave, time);
                // lineVerticesMap.set(id, vertices);
            }
        }
        return { lineVerticesMap, maskVerticesMap };
    }


    /**
     * @description 将wave 的值转化为渲染需要用到的值
     * @param {*} value 
     * @return {{
     *  y: number       // 裁剪空间的纵坐标
     *  color: number   // 颜色的索引 颜色rgb = gl_Colors[color]
     * }}
     */
    translateValue2RenderParameter(value) {
        switch (value) {
            case 0:             return { y: -1, color: 2 };       // 0 value
            case 1:             return { y: 1,  color: 3 };       // 1 value
            case 2: case 3:     return { y: -1, color: 4 };       // 不定态 x
            case 4: case 5:     return { y: 0,  color: 2 };       // 高阻态 z
            default:            return { y: -1, color: 7 };       // 其他，我也不知道还有啥
        }
    }

    /**
     * 
     * @param {{x: number, y: number, color: number} | undefined} p0 前一个点
     * @param {{x: number, y: number, color: number} | undefined} p1 当前的点
     * @param {{x: number, y: number, color: number} | undefined} p2 后一个点
     * @returns {number} 这是 widthshift 的索引，只需要 + 4 再 % 8 就能得到另一个 
     */
    makeWidthShiftIndexByPoints(p0, p1, p2) {
        if (p0 === undefined) {
            if (p1.y === p2.y) {
                return 0;
            } else if (p1.x === p2.x) {
                return 6;
            }
        } else if (p2 === undefined) {
            if (p1.y === p0.y) {
                return 0;
            } else if (p1.x === p0.x) {
                return 6;
            }
        } else {
            if (p0.x !== p1.x && p0.y === p1.y && p1.x === p2.x && p1.y !== p2.y) {
                if (p2.y > p1.y) {
                    return 1;
                } else {
                    return 7;
                }
            } else if (p0.x === p1.x && p0.y !== p1.y && p1.x !== p2.x && p1.y === p2.y) {
                if (p1.y > p0.y) {
                    return 1;
                } else {
                    return 7;
                }
            }
        }
    }

    /**
     * 
     * @param {number} x1 
     * @param {number} y1 
     * @param {number} x2 
     * @param {number} y2 
     * @param {number} color 
     * @param {number} wsIndex
     * @returns {number[]} 
     */
    makeRectangleVertices(x1, y1, x2, y2, color, wsIndex = 0) {
        const r1 = [x1, gl_Shifts_map.get(y2), color, wsIndex];
        const r2 = [x1, gl_Shifts_map.get(y1), color, wsIndex];
        const r3 = [x2, gl_Shifts_map.get(y1), color, wsIndex];
        const r4 = [x2, gl_Shifts_map.get(y2), color, wsIndex];
        return [
            ...r1, ...r2, ...r3,
            ...r1, ...r3, ...r4, 
        ]
    }

    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {{
     *  lineVertices: Uint32Array
     *  maskVertices: Uint32Array
     * }}
     */
    makeBitVertex(wave, time, debug = false) {
        const length = wave.length;
        // 先将节点数据转化为裁剪空间的坐标
        const perspectivePoints = [];

        for (let i = 0; i < length; ++ i) {
            // const currentWave = wave[(i === 0) ? 0 : (i - 1)];
            const currentWave = wave[i];
            const nextWave = (i === (length - 1)) ? wave[i] : wave[i + 1];

            const t1 = currentWave[0];
            const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
            const value1 = currentWave[1];
            const value2 = nextWave[1];
            
            const renderParam1 = this.translateValue2RenderParameter(value1);
            const renderParam2 = this.translateValue2RenderParameter(value2);

            if (i === 0) {
                perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
                perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
            } else {
                const lastPoint = perspectivePoints.at(-1);
                if ((lastPoint.y !== renderParam1.y) || (lastPoint.color !== renderParam1.color)) {
                    perspectivePoints.push({ x: t1, y: renderParam1.y, color: renderParam1.color });
                }
                if ((renderParam1.y !== renderParam2.y) || (renderParam1.color !== renderParam2.color)) {
                    perspectivePoints.push({ x: t2, y: renderParam1.y, color: renderParam1.color });
                }
            }
        }

        // 确保最后一个点延申到了 time
        const lastPoint = perspectivePoints.at(-1);
        if (lastPoint.x < time) {
            perspectivePoints.push({ x: time, y: lastPoint.y, color: lastPoint.color });
        }

        // 计算出传入 shader 四元组数组
        // 四元组： (x, yshift_index, color_index, width_shift_index)
        const pointNum = perspectivePoints.length;
        const lineVertices = [];
        const maskVertices = [];

        // const lineVertices = [
        //     0, 0, 2, 0,
        //     10, 0, 2, 0,
        //     10, 4, 2, 0,
        //     20, 4, 2, 0,
        //     20, 0, 2, 0,
        //     30, 0, 2, 0,
        //     30, 4, 2, 0
        // ];
        // return new Uint32Array(lineVertices);
        

        // 制作 lineVertices
        for (let i = 0; i < pointNum; ++ i) {
            // p0: 上一个点
            // p1: 当前的点
            // p2: 下一个点
            const p0 = perspectivePoints[i - 1];
            const p1 = perspectivePoints[i];
            const p2 = perspectivePoints[i + 1];
            // p1.x, y1Index, p1.color, 0,
            const y1Index = gl_Shifts_map.get(p1.y);

            const wsIndex = this.makeWidthShiftIndexByPoints(p0, p1, p2);
            lineVertices.push(
                p1.x, y1Index, p1.color, wsIndex,
                p1.x, y1Index, p1.color, (wsIndex + 4) % 8
            );

            // 防止颜色不同导致单个图元内出现两个颜色，这会引发shader的渐变
            if (p2 !== undefined && p1.color !== p2.color) {
                lineVertices.push(
                    p1.x, y1Index, p2.color, wsIndex,
                    p1.x, y1Index, p2.color, (wsIndex + 4) % 8
                );
            }
        }

        // 制作 maskVertices
        for (let i = 0; i < pointNum; ++ i) {
            const p1 = perspectivePoints[i];

            // 开头就有一条线
            if (i === 0 && p1.y === 1) {
                while (perspectivePoints[++ i] && perspectivePoints[i].y === 1);
                // 回退
                if (-- i > 0) {
                    // 四元组： (x, yshift_index, color_index, width_shift_index)
                    const rectangleVertices = this.makeRectangleVertices(p1.x, 1, perspectivePoints[i].x, -1, p1.color + 10, 4);
                    // 三角图元画矩形
                    maskVertices.push(...rectangleVertices);
                    continue;
                }
            }

            // 上升沿才使用
            const p2 = perspectivePoints[i + 1];
            const p3 = perspectivePoints[i + 2];

            if (p1 === undefined || p2 === undefined || p3 === undefined) {
                continue;
            }
            if (p2.y > p1.y) {
                // 矩形的四个点
                // 四元组： (x, yshift_index, color_index, width_shift_index)
                const rectangleVertices = this.makeRectangleVertices(p2.x, p2.y, p3.x, p1.y, p2.color + 10, 4);
                // 三角图元画矩形
                maskVertices.push(...rectangleVertices);
            }
        }


        if (debug) {
            console.log(perspectivePoints);
            console.log(pointNum);
            console.log(lineVertices);
        }

        const Uint32lineVertices = new Uint32Array(lineVertices);
        const Uint32maskVertices = new Uint32Array(maskVertices);

        return {
                lineVertices : Uint32lineVertices,
                maskVertices : Uint32maskVertices
            };
    }

    /**
     * 
     * @param {Array<string | number>} wave 
     * @param { number } time 
     * @returns {Uint32Array}
     */
    makeVecVertex(wave, time) {
        const vertices = [];
        const length = wave.length;
        for (let i = 0; i < length; ++ i) {
            const [t1, val, msk] = wave[i];
            const t2 = (i === (length - 1)) ? time : wave[i + 1][0];
            // t1(val) --- t2(val)

            if (msk) {
                vertices.push(
                    t1, 0, 0,
                    t2, 0, 0,
                    t2, 0, 4,
                    t2, 4, 4,
                    t1, 6, 4,
                    t1, 0, 4,
                    t1, 3, 4,
                    t2, 1, 4,
                    t2, 0, 4
                );
            } else {
                vertices.push(
                    t1, 0, 0,
                    t2, 0, 0,
                    t2, 0, 5,
                    t2, 4, 5,
                    t1, 6, 5,
                    t1, 0, 5,
                    t1, 3, 5,
                    t2, 1, 5,
                    t2, 0, 5
                );
            }

        }

        return new Uint32Array(vertices);
    }

    initData() {
        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;
        const globalLookup = this.globalLookup;
        const lineVerticesMap = this.lineVerticesMap;
        const maskVerticesMap = this.maskVerticesMap;
        
        for (const id of Reflect.ownKeys(globalLookup.chango)) {
            const signalItem = globalLookup.chango[id];

            const lineVertices = lineVerticesMap.get(id);
            if (lineVertices === undefined) {
                // console.warn(`无法找到 link 为 ${id} 的顶点数据`);
                continue;
            }
            
            // 创建并设置 绘制wave轮廓 主体轮廓的 缓冲区、vao、顶点设置
            const lineVertexBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, lineVertexBuffer);
            gl.bufferData(gl.ARRAY_BUFFER, lineVertices, gl.STATIC_DRAW);
            signalItem.lineVao = gl.createVertexArray();
            gl.bindVertexArray(signalItem.lineVao);
            gl.vertexAttribIPointer(webglLocation.pos, 4, gl.UNSIGNED_INT, 0, 0);
            gl.enableVertexAttribArray(webglLocation.pos);


            const maskVertices = maskVerticesMap.get(id);
            if (maskVertices === undefined) {
                continue;
            }

            // 创建并设置 绘制wave半透明遮罩层 主体轮廓的 缓冲区、vao、顶点设置
            const maskVertexBuffer = gl.createBuffer();
            gl.bindBuffer(gl.ARRAY_BUFFER, maskVertexBuffer);
            gl.bufferData(gl.ARRAY_BUFFER, maskVertices, gl.STATIC_DRAW);
            signalItem.maskVao = gl.createVertexArray();
            gl.bindVertexArray(signalItem.maskVao);
            gl.vertexAttribIPointer(webglLocation.pos, 4, gl.UNSIGNED_INT, 0, 0);
            gl.enableVertexAttribArray(webglLocation.pos);
        }

        gl.uniform4fv(webglLocation.colors, gl_Colors);
        gl.uniform2fv(webglLocation.widthShifts, gl_WidthShifts);
        gl.uniform2fv(webglLocation.shifts, gl_Shifts);
    }

    render() {
        if (this.animationHandler !== undefined) {
            cancelAnimationFrame(this.animationHandler);
        }
        
        const canvas = this.canvas;
        const webglLocation = this.webglLocation;
        const gl = webglLocation.gl;
        const globalLookup = this.globalLookup;
        const lineVerticesMap = this.lineVerticesMap;
        const maskVerticesMap = this.maskVerticesMap;
        const elements = this.elements;

        const timeScaleHeight = parseInt(document.body.style.getPropertyValue('--time-scale-height').match(/\d+/)[0]);
        const sidebarPadding = parseInt(document.body.style.getPropertyValue('--sidebar-padding').match(/\d+/)[0]);
        const vcdRenderPadding = parseInt(document.body.style.getPropertyValue('--vcd-render-padding').match(/\d+/)[0]);
        const canvasPaddingTop = timeScaleHeight + sidebarPadding - vcdRenderPadding;

        this.animationHandler = window.requestAnimationFrame(() => {
            const { width, height, xScale, xOffset, yOffset, yStep, yDuty } = this.pstate;
            const canvasHeight = height - canvasPaddingTop;
            const canvasWidth = width;

            // 默认 1594
            canvas.width = canvasWidth;

            // 默认 1260
            canvas.height = canvasHeight;

            // 设置 glsl 变量
            gl.uniform2f(webglLocation.scale,
                2 * xScale / canvasWidth,
                yStep * yDuty / canvasHeight
            );
            console.log(yStep, yDuty, canvasHeight);
            // 设置 webgl 和 canvas 大小位置一致
            gl.viewport(0, 0, canvasWidth, canvasHeight);
            // 清楚颜色缓冲区，也就是删除上一次的渲染结果
            gl.clear(gl.COLOR_BUFFER_BIT);


            // 根据 globalLookup 当前激活的需要渲染的信号进行渲染
            let index = 0;
            for (const signal of globalLookup.currentWires) {
                const wave = globalLookup.chango[signal.link].wave;
                // this.makeBitVertex(wave, globalLookup.time, true);
                const signalItem = globalLookup.chango[signal.link];
                if (!signalItem) {
                    return;
                }

                // TODO: 将此处的 offset 计算中的参数和 globalSetting 的数字形成关联
                gl.uniform2f(webglLocation.offset, 
                    (2 * xOffset / width) - 1, 
                    (2 * yOffset - 2 * yStep * (index + .7)) / canvasHeight + 1
                );
                

                console.log('offset y', (2 * yOffset - 2 * yStep * (index + .7)) / canvasHeight + 1);

                // 根据 lineVao 进行绘制
                const lineVertices = lineVerticesMap.get(signal.link);
                const maskVertices = maskVerticesMap.get(signal.link);
                
                if (signal.size === 1) {
                    // 如果是 width 为 1 的
                    gl.bindVertexArray(signalItem.lineVao);
                    gl.drawArrays(gl.TRIANGLE_STRIP, 0, lineVertices.length / 4);
``
                    gl.bindVertexArray(signalItem.maskVao);
                    gl.drawArrays(gl.TRIANGLES, 0, maskVertices.length / 4);
                } else {
                    // 如果是 width 大于 1 的
                    gl.drawArrays(gl.LINE_STRIP, 0, lineVertices.length / 4);
                }

                index ++;
            }

            this.plugins.map(fn => fn(globalLookup, this.pstate, elements));
            this.animationHandler = undefined;
        });
    }
}

module.exports = WebGL2WaveRender;